#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""DCT demo
"""
__author__ = "Miguel Hernández Cabronero <mhernandez@deic.uab.cat>"
__date__ = "03/09/2019"

import sys
import os
import math
import collections
import numpy as np
import scipy.fftpack
from matplotlib import pyplot as plt
import cv2
import glob
import random
import itertools
import pandas

pretty_dict = {
    "mse": "MSE",
    "dct_entropy": "Entropy after DCT",
    "no_quantization": "No quantization",
    "only_h_1": "Only DC and DCT[1,0]",
    "only_h_2": "Only DC and DCT[2,0]",
}
pretty_dict.update({f"jpeg_Q{Q}": f"JPEG Q{Q}" for Q in range(101)})


def dct(block, B=8):
    """:return the 2D DCT-II (as in JPG) of the block
    """
    block = np.array(block, dtype=float)
    dc = block.mean()
    block = block - dc
    # block -= 2 ** (B - 1)
    dct = scipy.fftpack.dct(scipy.fftpack.dct(block, axis=1, norm="ortho"), axis=0, norm="ortho")
    dct[0, 0] = dc
    return dct
    # Roughly equivalent to
    # dct_block = np.zeros(block.shape)
    # for u in range(block.shape[1]):
    #     alpha_u = (1 / math.sqrt(2)) if u == 0 else 1
    #     for v in range(block.shape[0]):
    #         alpha_v = (1 / math.sqrt(2)) if v == 0 else 1
    #         dct_block[v, u] = 0.25 * alpha_u * alpha_v * sum(
    #             block[x, y]
    #             * math.cos((2 * x + 1) * u * math.pi / 16)
    #             * math.cos((2 * y + 1) * v * math.pi / 16)
    #             for x in range(block.shape[0])
    #             for y in range(block.shape[1]))
    # return dct_block


def inverse_dct(dct_block, B=8):
    dct_block = np.array(dct_block)
    dc = dct_block[0, 0]
    dct_block[0, 0] = 0
    block = scipy.fftpack.idct(scipy.fftpack.idct(dct_block, axis=0, norm="ortho"), axis=1, norm="ortho")
    block += dc
    return block
    # Roughly equivalent to
    # for x in range(block.shape[1]):
    #     for y in range(block.shape[0]):
    #         block[x, y] = 0.25 * sum(
    #             (1 / math.sqrt(2) if u == 0 else 1)
    #             * (1 / math.sqrt(2) if v == 0 else 1)
    #             * dct_block[v, u]
    #             * math.cos((2 * x + 1) * u * math.pi / 16)
    #             * math.cos((2 * y + 1) * v * math.pi / 16)
    #             for u in range(dct_block.shape[1])
    #             for v in range(dct_block.shape[0]))
    # return block + 2 ** (B - 1)


def entropy(data):
    counter = collections.Counter(np.array(data, copy=False).flatten())
    probabilities = (v / data.size for k, v in counter.items())
    return - sum(p * math.log2(p) for p in probabilities)


def display_dct(block):
    assert block.shape == (8, 8)
    B = 8
    transformed_matrix = dct(block, B=B)
    reconstructed_matrix = inverse_dct(transformed_matrix, B=B).round().astype(int)

    plt.figure()
    plt.subplot(321)
    plt.title("Original matrix")
    plt.imshow(block, cmap=plt.cm.gray)
    plt.axis("off")
    plt.subplot(323)
    plt.title("Transformed matrix")
    plt.imshow(transformed_matrix, cmap=plt.cm.gray)
    plt.axis("off")
    plt.subplot(324)
    plt.title("sign(Transformed matrix)")
    plt.imshow(np.sign(transformed_matrix), cmap=plt.cm.gray)
    plt.axis("off")
    plt.subplot(325)
    plt.title("|Transformed matrix|")
    plt.imshow(np.abs(transformed_matrix), cmap=plt.cm.gray)
    plt.axis("off")
    plt.subplot(326)
    plt.title("|Transformed matrix (no DC)|")
    no_dc = np.abs(transformed_matrix)
    no_dc[0, 0] = 0
    plt.imshow(no_dc, cmap=plt.cm.gray)
    plt.axis("off")
    plt.show()
    plt.close()


selected_x = None
selected_y = None


def interactive_dct(image_path=None):
    global selected_x
    global selected_y

    while True:
        if image_path is None:
            input_image = cv2.imread(random.sample(glob.glob(os.path.join("img", "*.jpg")), 1)[0])
        else:
            input_image = cv2.imread(image_path)

        input_image[:, :, 0], input_image[:, :, 1], input_image[:, :, 2] = \
            np.array(input_image[:, :, 2]), \
            np.array(input_image[:, :, 1]), \
            np.array(input_image[:, :, 0])

        selected_x, selected_y = None, None

        def onclick(event):
            global selected_x
            global selected_y
            if event.button == 1:
                selected_x, selected_y = int(event.xdata), int(event.ydata)
                plt.close()
                return selected_x, selected_y
            elif event.button == 3:
                sys.exit(0)

        while selected_x is None or selected_y is None:
            fig = plt.figure()
            plt.imshow(input_image)
            plt.title("Left click to select block - right click to exit")
            fig.canvas.mpl_connect('button_press_event', onclick)
            plt.show()

        block = input_image[selected_y - 4:selected_y + 4, selected_x - 4:selected_x + 4, 1]
        display_dct(block)
        if image_path is not None:
            break


def generate_impulses(val):
    os.makedirs("impulse", exist_ok=True)
    for x in range(8):
        for y in range(8):
            m = np.zeros((8, 8))
            m[y, x] = val
            cv2.imwrite(f"impulse/impulse_x{x}_y{y}.png", m)
            idct_m = np.clip(inverse_dct(m) + 128, 0, 255)
            cv2.imwrite(f"impulse/response_x{x}_y{y}.png", idct_m)


def get_jpeg_qmatrix(Q):
    Tb = np.array([
        [16, 11, 10, 16, 24, 40, 51, 61],
        [12, 12, 14, 19, 26, 58, 60, 55],
        [14, 13, 16, 24, 40, 57, 69, 56],
        [14, 17, 22, 29, 51, 87, 80, 62],
        [18, 22, 37, 56, 68, 109, 103, 77],
        [24, 35, 55, 64, 81, 104, 113, 92],
        [49, 64, 78, 87, 103, 121, 120, 101],
        [72, 92, 95, 98, 112, 100, 103, 99]])
    if Q < 50:
        S = 5000 / Q
    else:
        S = 200 - 2 * Q

    qmatrix = np.floor((S * Tb + 50) / 100).astype(int)
    qmatrix[qmatrix == 0] = 1
    return qmatrix


def zoom(img, zoom_factor=6):
    zoom_factor = 6
    zwidth, zheight = img.shape[0] // zoom_factor, img.shape[1] // zoom_factor
    img = img[img.shape[0] // 2 - zwidth // 2:img.shape[0] // 2 - zwidth // 2 + zwidth,
          img.shape[1] // 2 - zheight // 2:img.shape[1] // 2 - zheight // 2 + zheight]
    return cv2.resize(img, (0, 0), fx=zoom_factor, fy=zoom_factor, interpolation=cv2.INTER_NEAREST)


def filter_dct():
    output_dir = "filter_panels"
    os.makedirs(output_dir, exist_ok=True)

    Q_values = [int(Q) for Q in np.linspace(1, 100, 11)]
    jpg_matrices = [get_jpeg_qmatrix(Q=int(Q)) for Q in Q_values]
    jpg_labels = [f"jpeg_Q{Q}" for Q in Q_values]
    only_h_matrices = []
    only_v_matrices = []
    only_hv_matrices = []
    only_h_labels = []
    only_v_labels = []
    only_hv_labels = []
    for i in range(4):
        only_h_matrices.append(np.zeros((8, 8)) + 10000)
        only_h_labels.append(f"only_h_{i}")
        only_h_matrices[-1][i, 0] = 1
        only_v_matrices.append(np.zeros((8, 8)) + 10000)
        only_v_labels.append(f"only_v_{i}")
        only_v_matrices[-1][0, i] = 1
        only_hv_matrices.append(np.zeros((8, 8)) + 10000)
        only_hv_labels.append(f"only_hv_{i}")
        only_hv_matrices[-1][i, i] = 1
    for m in itertools.chain(only_h_matrices, only_v_matrices, only_hv_matrices):
        m[0,0] = 1

    entropy_df = pandas.DataFrame(columns=["image", "label", "original_entropy", "dct_entropy", "reconstructed_entropy", "mse"])
    # for filter_matrix, label in [(full_quality_matrix, "full_quality"),
    #                              (dconly_matrix, "dc_only")
    #                              ] + \
    #                             [(get_jpeg_qmatrix(Q=Q), f"jpegQ{Q}") for Q in [100, 50]]:
    for filter_matrix, label in itertools.chain(
            [(np.zeros((8,8)) + 1, "no_quantization")],
            zip(jpg_matrices, jpg_labels),
            zip(only_h_matrices, only_h_labels),
            zip(only_v_matrices, only_v_labels),
            zip(only_hv_matrices, only_hv_labels)):
        print()
        print(f"label = {label}")

        filter_matrix = filter_matrix.astype(int)
        for image_path in glob.glob("img/*jpg"):
            output_path_full = os.path.join(output_dir,
                                            f"dct_filter_{label}_{os.path.basename(image_path).replace('.jpg', '.png')}")
            output_path_zoom = os.path.join(output_dir,
                                            f"zoom_dct_filter_{label}_{os.path.basename(image_path).replace('.jpg', '.png')}")

            img = cv2.imread(image_path)
            img = np.dstack([img[:, :, c] for c in reversed(range(3))])
            original_img = np.array(img)
            big_panel = np.array(img)
            zoom_panel = zoom(img)

            dct_img = np.zeros(img.shape)
            x = 0
            while 8 * (x + 1) < img.shape[0]:
                y = 0
                while 8 * (y + 1) < img.shape[1]:
                    for c in range(3):
                        # forward DCT
                        block = np.array(img[8 * x:8 * (x + 1), 8 * y:8 * (y + 1), c])
                        # fwd DCT
                        dct_block = dct(block)
                        # quantization
                        sign = np.sign(dct_block)
                        dct_block = sign * filter_matrix * (np.abs(dct_block).round().astype(int) // filter_matrix)
                        dct_img[8 * x:8 * (x + 1), 8 * y:8 * (y + 1), c] = dct_block
                        # inverse DCT
                        reconstructed_block = inverse_dct(dct_block).round().astype(int)
                        img[8 * x:8 * (x + 1), 8 * y:8 * (y + 1), c] = np.clip(reconstructed_block, 0, 255)
                    y += 1
                x += 1

            entropy_df.loc[len(entropy_df)] = pandas.Series({
                "image": image_path,
                "label": label,
                "original_entropy": entropy(original_img),
                "dct_entropy": entropy(dct_block),
                "reconstructed_entropy": entropy(img),
                "mse": ((img-original_img)**2).mean(),
            })
            entropy_df.to_csv("dct_entropies.csv")

            big_panel = np.concatenate([big_panel, img, np.abs(original_img - img)], axis=1)

            zoom_panel = np.concatenate([zoom_panel, zoom(img), zoom(np.abs(original_img - img))], axis=1)
            cv2.imwrite(
                output_path_full,
                np.dstack([big_panel[:, :, c] for c in reversed(range(3))]))
            cv2.imwrite(
                output_path_zoom,
                np.dstack([zoom_panel[:, :, c] for c in reversed(range(3))]))

            print(os.path.basename(image_path))
            print(f"Entropies:")
            print(f"original: {entropy(original_img)}")
            print(f"{label}: {entropy(img)}")
    entropy_df.to_csv("dct_entropies.csv")

def render_jpg_matrices():
    os.makedirs("jpeg_matrices", exist_ok=True)
    for Q in range(10, 100, 9):
        cv2.imwrite(f"quantization_matrices/jpeg_Q{Q}.png", cv2.resize(
            get_jpeg_qmatrix(Q), (0, 0), fx=100, fy=100, interpolation=cv2.INTER_NEAREST))

def analyze_entropies():
    plt.figure()
    plt.xlabel(pretty_dict["dct_entropy"])
    plt.ylabel(pretty_dict["mse"])
    colors = ["r", "g", "b", "orange", "gray"]
    markers = ["o", "x", "+"]
    used_labels = set()

    selected_labels = ["no_quantization", "only_h_1", "only_h_2"] + [f"jpeg_Q{Q}" for Q in range(10, 101, 10)]

    df = pandas.read_csv("dct_entropies.csv")
    for j, (label, group_df) in enumerate(df.groupby("label")):
        if label not in selected_labels:
            continue
        plt.scatter(group_df["dct_entropy"].mean(), group_df["mse"].mean(),
                    label=pretty_dict[label] if label in pretty_dict else label,
                    color=colors[j % len(colors)],
                    marker=markers[j % len(markers)],
                    s=50)
    plt.legend()
    plt.tight_layout()
    plt.savefig(f"dctentropy_mse.pdf")
    plt.close()

if __name__ == '__main__':
    # generate_impulses(200)
     interactive_dct(sys.argv[1] if len(sys.argv) > 1 else None)
    # filter_dct()
    #analyze_entropies()

    # render_jpg_matrices()
